from mellifera.executors.trio import TrioExecutor
try:
    from mellifera.executors.nsmainthread import NSMainThreadExecutor
    HAS_NSMAINTHREAD = True
except ModuleNotFoundError:
    HAS_NSMAINTHREAD = False



from __future__ import annotations

from abc import ABC, abstractmethod
from types import SimpleNamespace
import functools
import logging

from typing import TYPE_CHECKING, Any, Optional
import threading

from mellifera.executor import Executor
from mellifera.services import TrioService, NSMainThreadService
from mellifera.service import ServiceState
import trio.lowlevel

if TYPE_CHECKING:
    from mellifera.service import Service


class Orchestrator:

    def __init__(self) -> None:
        self.logger = logging.getLogger("calsiprovis.orchestrator.Orchestrator")
        self.services = dict()
        self.system_nursery = None
        self.service_nursery = None
        self.trio_token = None
        self.running = False
        self.to_run = []

        self.write_channel, self.recv_channel = trio.open_memory_channel(1000)

        from mellifera.executors import TrioExecutor, HAS_NSMAINTHREAD
        if HAS_NSMAINTHREAD:
            from mellifera.executors import NSMainThreadExecutor

        self.trio = TrioExecutor(self)
        if HAS_NSMAINTHREAD:
            self.mainthread_executor = NSMainThreadExecutor(self)
        else:
            self.mainthread_executor = None

    def get_service(self, name: str) -> Service:
        if name in self.services:
            return self.services[name]
        else:
            raise ValueError(f"Service with name `{name}` not found")

    def register_service(self, service: Service, name: str) -> None:
        match service:
            case TrioService():
                service.register(self, self.trio, name)
            case NSMainThreadService():
                if not self.mainthread_executor:
                    raise ValueError("Cannot have a main_thread_service without a main thread executor")
                service.register(self, self.mainthread_executor, name)
                self.mainthread_executor.service = service
            case _:
                raise ValueError(f"Cannot handle service of type {type(service)}")
        self.services[name] = service

    def start_service(self, service: Service | str) -> None:
        if isinstance(service, str):
            service = self.get_service(service)
        match service:
            case TrioService():
                if service.state == ServiceState.REGISTERED:
                    service.state = ServiceState.WILL_BE_STARTED
                    if self.service_nursery:
                        self.service_nursery.start_soon(self.trio.run_service, service)
                    else:
                        self.to_run.append(service)
            case NSMainThreadService():
                if service.state == ServiceState.REGISTERED:
                    if not self.mainthread_executor:
                        raise ValueError("Cannot have a main_thread_service without a main thread executor")
                    service.state = ServiceState.WILL_BE_STARTED
                    self.mainthread_executor.start_service(service)
            case _:
                raise ValueError(f"Cannot start service of type {type(service)}")

    def stop_service(self, service: Service | str) -> None:
        if isinstance(service, str):
            service = self.get_service(service)
        if service.state.value < ServiceState.STOPPING.value:
            match service:
                case TrioService():
                    self.service_nursery.start_soon(self.trio.stop_service, service)
                case NSMainThreadService():
                    if self.mainthread_executor and self.mainthread_executor.service == service:
                        self.mainthread_executor.stop()
                    else:
                        raise ValueError(f"Cannot start service of type {type(service)}, mainthread_executor not active")
                case _:
                    raise ValueError(f"Cannot start service of type {type(service)}")

    def enable_service(self, service: Service, name: str) -> None:
        self.register_service(service, name)
        self.start_service(service)

    def run_async_function(self, async_function):
        if not self.trio_token:
            raise ValueError("Trying to run async function, but trio is not started yet")
        async def closure():
            await self.write_channel.send(async_function)
        trio.from_thread.run(closure, trio_token=self.trio_token)

    async def wait_for_mainthread(self):
        await trio.sleep(0)
        if self.mainthread_executor and self.mainthread_executor.service:
            await self.mainthread_executor.service.shutdown
            self.stop_all()

    async def process_messages(self):
        with self.recv_channel:
            async with trio.open_nursery() as nursery:
                nursery.start_soon(self.wait_for_mainthread)
                async for async_function in self.recv_channel:
                    async def closure():
                        try:
                            await async_function()
                        except Exception:
                            self.logger.error("Wild async function raised an error", exc_info=True)
                    nursery.start_soon(closure)

    def stop(self):
        self.write_channel.close()

    def run(self):
        self.logger.info('Orchestrator.run starting')
        if self.mainthread_executor:
            self.start_sync()
            self.mainthread_executor.run_sync()
        else:
            self.run_sync()
        self.logger.info('Orchestrator.run ending')

    def start_sync(self):
        self.logger.debug("Starting in sync mode")
        self.thread = threading.Thread(target=trio.run, args=(self.run_async,), daemon=False)
        self.thread.start()

    def run_sync(self):
        trio.run(self.run_async)

    async def run_async(self):
        self.thread = threading.current_thread()
        self.trio_token = trio.lowlevel.current_trio_token()
        self.logger.debug('run started')
        async with trio.open_nursery() as system_nursery:
            self.logger.debug('nursery opened')
            self.system_nursery = system_nursery
            system_nursery.start_soon(self.process_messages)
            if self.mainthread_executor:
                self.mainthread_executor.trio_startup.set()
            await trio.sleep(1) # give mainthread_executor some time to startup
            async with trio.open_nursery() as service_nursery:
                self.service_nursery = service_nursery
                service_nursery.start_soon(self.wait_for_mainthread)
                for service in self.to_run:
                    self.service_nursery.start_soon(self.trio.run_service, service)
                self.to_run = []
            self.stop()
        self.logger.info("Nursery of mellifera.executors.TrioExecutor.run closed")

    def stop_all(self):
        for service in self.services.values():
            self.stop_service(service)



from mellifera.executors import TrioExecutor, HAS_NSMAINTHREAD

if HAS_NSMAINTHREAD:
    from mellifera.executors import NSMainThreadExecutor

from mellifera.orchestrator import Orchestrator
from mellifera.service import Service, expose
from mellifera.services import TrioService, NSMainThreadService



from typing import TYPE_CHECKING

import trio
import trio.lowlevel
import threading
import inspect
import logging
from queue import Queue
from types import SimpleNamespace

from mellifera.services.trio import TrioService, ServiceState
from mellifera.orchestrator import Orchestrator
from mellifera.service import Service
from mellifera.executor import Executor


class TrioExecutor(Executor):
    requires_run = False

    def __init__(self, orchestrator) -> None:
        self.orchestrator = orchestrator
        self.services = []
        self.logger = logging.getLogger("mellifera.executors.TrioExecutor")

        self.nursery = None
        self.to_start = []
        self.stopped = trio.Event()

    def run_exposed(self, f, service, *args, **kwargs):
        if threading.current_thread() == self.orchestrator.thread:
            return f(service, *args, **kwargs)
        else:
            if inspect.iscoroutinefunction(f):
                async def closure():
                    return await f(service, *args, **kwargs)
            else:
                async def closure():
                    return f(service, *args, **kwargs)
            self.orchestrator.run_async_function(closure)

    async def stop_service(self, service):
        try:
            with trio.fail_after(5):
                await service.stop()
                await service.stopped
        except trio.TooSlowError:
            with trio.move_on_after(10):
                await service.cancel()

    async def _run_service(self, service):
        try:
            service.construct()
        except Exception as e:
            self.logger.error(f"Error during construction of service {service.name}", exc_info=True)
            self.orchestrator.stop_all()
            raise e

        try:
            self.logger.debug(f'Initializing service {service.name}')
            await service.init()
            self.logger.debug(f'Initializing service {service.name} DONE')
        except Exception as e:
            self.logger.error(f"Error during initialisation of service {service.name}", exc_info=True)
            self.orchestrator.stop_all()
            raise e

        try:
            self.logger.debug(f'Running service {service.name}')
            await service.run()
            self.logger.debug(f'Running service {service.name} DONE')
        except Exception as e:
            self.logger.error(f"Error during run of service {service.name}", exc_info=True)
            self.orchestrator.stop_all()
            raise e

    async def run_service(self, service):
        try:
            async with trio.open_nursery() as nursery:
                service.nursery = nursery
                await self._run_service(service)
        finally:
            self.logger.debug(f"Finalizing {service.name}")
            with trio.CancelScope(shield=True):
                await service.finalize()
            self.logger.debug(f"Finalizing {service.name} done")



try:
    from Foundation import NSThread
    from libdispatch import dispatch_async, dispatch_sync, dispatch_get_main_queue
except ModuleNotFoundError:
    raise ModuleNotFoundError("To use mellifera.orchestrators.nsmainthread you need to have pyobjc installed and run on macos")

import signal

from types import SimpleNamespace
import logging
import threading
import functools

from mellifera.orchestrator import Orchestrator
from mellifera.services.nsmainthread import NSMainThreadService
from mellifera.service import Service
from mellifera.executor import Executor

from mellifera.service import ServiceState

def threadsafe(f):
    @functools.wraps(f)
    def inner(self, *args, **kwargs):
        return self.run_threadsafe(f, self, *args, **kwargs)

    return inner

class NSMainThreadExecutor(Executor):
    requires_run = True

    def __init__(self, orchestrator) -> None:
        self.orchestrator = orchestrator
        self.service = None
        self.logger = logging.getLogger("mellifera.executors.NSMainThreadExecutor")
        self.thread = None
        self.trio_startup = threading.Event()
        self._start_service = False

    def run_threadsafe(self, f, *args, **kwargs):
        self.logger.info(f"Running threadsafe: {f.__name__}")
        if self.thread is None or self.thread == threading.current_thread():
            r = f(*args, **kwargs)
            return None
        else:
            def closure():
                self.logger.debug(f'excecuted closure of {f.__name__} threadsafe via NSMainThreadExecutor')
                return f(*args, **kwargs)
            queue = dispatch_get_main_queue()
            r = dispatch_async(queue, closure)
            return None

    def run_exposed(self, f, service, *args, **kwargs):
        return self.run_threadsafe(f, service, *args, **kwargs)

    def start_service(self, service: Service) -> None:
        assert isinstance(service, NSMainThreadService)
        if self.service is None or self.service == service:
            self._start_service = True
        else:
            raise ValueError(f"Asked to start {service}, but registered {self.service}")

    @threadsafe
    def stop(self) -> None:
        self.service.stop()

    def wait_for_trio_startup(self) -> None:
        self.trio_startup.wait()

    def run_sync(self) -> None:
        self.logger.info("run_sync")
        self.thread = threading.current_thread()
        self.logger.info("Waiting for trio to startup")
        self.wait_for_trio_startup()
        self.logger.info("Trio started up, starting service")
        try:
            self.logger.debug("Starting up")
            if self.service and self._start_service:
                self.logger.debug("Starting service")
                try:
                    try:
                        self.logger.debug("Constructing service")
                        self.service.construct()
                        self.logger.debug("Initializing service")
                        self.service.init_sync()
                        self.logger.debug("Running service")
                        self.service.run_sync()
                        self.logger.debug("Running service done, stopping")
                    except Exception as e:
                        self.logger.error("Encountered an error", exc_info=True)
                        raise e
                    finally:
                        self.logger.error("Stopping all")
                finally:
                    self.service.finalize_sync()
            else:
                if not self.service:
                    self.logger.debug("No service, aborting run")
                elif not self._start_service:
                    self.logger.debug("I have a service, but startup was never requested. Not running service")

        finally:
            self.running = False
            self.logger.info("run_sync ended")



from abc import ABC, abstractmethod

from mellifera.service import Service

class Executor(ABC):

    @abstractmethod
    def run_exposed(self, f, service, *args, **kwargs):
        """Run function f in service with args and kwargs in a threadsafe matter

        This function is threadsafe.
        """
        ...
